@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# coding: utf-8
 
 import itertools
 import os
@@ -11,22 +10,35 @@ from glob import glob
 from xml.etree import ElementTree as stdlibElementTree
 
 import cryptography.exceptions
-from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives.asymmetric import dsa, ec, rsa
 from lxml import etree
 
+from signxml.signer import XMLSignatureReference
+
 sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
-from signxml import (  # noqa
+from signxml import (  # noqa:E402
+    CanonicalizationMethod,
+    DigestAlgorithm,
     InvalidCertificate,
     InvalidDigest,
     InvalidInput,
     InvalidSignature,
+    SignatureConstructionMethod,
+    SignatureMethod,
+    VerifyResult,
     XMLSignatureProcessor,
     XMLSigner,
     XMLVerifier,
     methods,
     namespaces,
 )
+from signxml.xades import (  # noqa:E402
+    XAdESDataObjectFormat,
+    XAdESSignaturePolicy,
+    XAdESSigner,
+    XAdESVerifier,
+    XAdESVerifyResult,
+)
 
 
 def reset_tree(t, method):
@@ -39,7 +51,7 @@ def reset_tree(t, method):
 
 class URIResolver(etree.Resolver):
     def resolve(self, url, id, context):
-        print("Resolving URL '%s'" % url)
+        print(f"Resolving URL '{url}'")
         return None
 
 
@@ -49,8 +61,18 @@ parser.resolvers.add(URIResolver())
 interop_dir = os.path.join(os.path.dirname(__file__), "interop")
 
 
-class TestVerifyXML(unittest.TestCase):
+class LoadExampleKeys:
+    def load_example_keys(self):
+        with open(os.path.join(os.path.dirname(__file__), "example.pem"), "rb") as fh:
+            crt = fh.read()
+        with open(os.path.join(os.path.dirname(__file__), "example.key"), "rb") as fh:
+            key = fh.read()
+        return crt, key
+
+
+class TestVerifyXML(unittest.TestCase, LoadExampleKeys):
     def test_example_multi(self):
+        cert, _ = self.load_example_keys()
         with open(os.path.join(os.path.dirname(__file__), "example.pem")) as fh:
             cert = fh.read()
         example_file = os.path.join(os.path.dirname(__file__), "example-125.xml")
@@ -61,7 +83,7 @@ class TestVerifyXML(unittest.TestCase):
         )
 
 
-class TestSignXML(unittest.TestCase):
+class TestSignXML(unittest.TestCase, LoadExampleKeys):
     def setUp(self):
         self.example_xml_files = (
             os.path.join(os.path.dirname(__file__), "example.xml"),
@@ -69,17 +91,19 @@ class TestSignXML(unittest.TestCase):
         )
         self.keys = dict(
             hmac=b"secret",
-            rsa=rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend()),
-            dsa=dsa.generate_private_key(key_size=1024, backend=default_backend()),
-            ecdsa=ec.generate_private_key(curve=ec.SECP384R1(), backend=default_backend()),
+            rsa=rsa.generate_private_key(public_exponent=65537, key_size=2048),
+            dsa=dsa.generate_private_key(key_size=1024),
+            ecdsa=ec.generate_private_key(curve=ec.SECP384R1()),
         )
 
     def test_basic_signxml_statements(self):
-        with self.assertRaisesRegex(InvalidInput, "Unknown signature method"):
+        self.assertEqual(SignatureMethod.RSA_SHA256.value, "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256")
+        self.assertEqual(SignatureConstructionMethod.enveloped, methods.enveloped)
+        with self.assertRaisesRegex(InvalidInput, "Unknown signature construction method"):
             signer = XMLSigner(method=None)
 
         with self.assertRaisesRegex(InvalidInput, "must be an XML element"):
-            XMLSigner().sign("x")
+            XMLSigner(signature_algorithm="hmac-sha256").sign("x", key=b"abc")
 
         digest_algs = {"sha1", "sha224", "sha256", "sha384", "sha512"}
         sig_algs = {"hmac", "dsa", "rsa", "ecdsa"}
@@ -87,7 +111,6 @@ class TestSignXML(unittest.TestCase):
         c14n_algs = {
             "http://www.w3.org/2001/10/xml-exc-c14n#",
             "http://www.w3.org/2001/10/xml-exc-c14n#WithComments",
-            XMLSignatureProcessor.default_c14n_algorithm,
         }
 
         all_methods = itertools.product(digest_algs, sig_algs, hash_algs, methods, c14n_algs)
@@ -127,15 +150,18 @@ class TestSignXML(unittest.TestCase):
                 signed_data = etree.tostring(signed)
                 XMLVerifier().verify(signed_data, **verify_kwargs)
                 XMLVerifier().verify(signed_data, parser=parser, **verify_kwargs)
-                (_d, _x, _s) = XMLVerifier().verify(signed_data, id_attribute="Id", **verify_kwargs)
+                res = XMLVerifier().verify(signed_data, id_attribute="Id", **verify_kwargs)
+                self.assertIsInstance(res, VerifyResult)
+                for attr in "signed_data", "signed_xml", "signature_xml":
+                    self.assertTrue(hasattr(res, attr))
 
-                if _x is not None:
+                if res.signed_xml is not None:
                     # Ensure the signature is not part of the signed data
-                    self.assertIsNone(_x.find(".//{http://www.w3.org/2000/09/xmldsig#}Signature"))
-                    self.assertNotEqual(_x.tag, "{http://www.w3.org/2000/09/xmldsig#}Signature")
+                    self.assertIsNone(res.signed_xml.find(".//{http://www.w3.org/2000/09/xmldsig#}Signature"))
+                    self.assertNotEqual(res.signed_xml.tag, "{http://www.w3.org/2000/09/xmldsig#}Signature")
 
                 # Ensure the signature was returned
-                self.assertEqual(_s.tag, "{http://www.w3.org/2000/09/xmldsig#}Signature")
+                self.assertEqual(res.signature_xml.tag, "{http://www.w3.org/2000/09/xmldsig#}Signature")
 
                 if method == methods.enveloping:
                     with self.assertRaisesRegex(InvalidInput, "Unable to resolve reference URI"):
@@ -177,10 +203,7 @@ class TestSignXML(unittest.TestCase):
 
         tree = etree.parse(self.example_xml_files[0])
         ca_pem_file = os.path.join(os.path.dirname(__file__), "example-ca.pem").encode("utf-8")
-        with open(os.path.join(os.path.dirname(__file__), "example.pem"), "rb") as fh:
-            crt = fh.read()
-        with open(os.path.join(os.path.dirname(__file__), "example.key"), "rb") as fh:
-            key = fh.read()
+        crt, key = self.load_example_keys()
         for hash_alg in "sha1", "sha256":
             for method in methods.enveloped, methods.enveloping:
                 print(hash_alg, method)
@@ -218,7 +241,7 @@ class TestSignXML(unittest.TestCase):
             from OpenSSL.crypto import X509
 
             with open(os.path.join(interop_dir, "TR2012", "rsa-cert.der"), "rb") as fh:
-                return [X509.from_cryptography(load_der_x509_certificate(fh.read(), backend=default_backend()))]
+                return [X509.from_cryptography(load_der_x509_certificate(fh.read()))]
 
         signature_files = glob(os.path.join(interop_dir, "TR2012", "signature*.xml"))
         for signature_file in signature_files:
@@ -372,16 +395,16 @@ class TestSignXML(unittest.TestCase):
                     elif "TR2012" not in signature_file:
                         raise
 
-    def test_signxml_changing_signature_namespace_prefix(self):
+    def test_changing_signature_namespace_prefix(self):
         data = etree.parse(self.example_xml_files[0]).getroot()
         signer = XMLSigner()
         signer.namespaces = dict(digi_sign=namespaces["ds"])
         signed = signer.sign(data, key=self.keys["rsa"])
         signed_data = etree.tostring(signed)
-        expected_match = ("<digi_sign:Signature xmlns:" 'digi_sign="%s">') % namespaces["ds"]
+        expected_match = f'<digi_sign:Signature xmlns:digi_sign="{namespaces["ds"]}">'
         self.assertTrue(re.search(expected_match.encode("ascii"), signed_data))
 
-    def test_signxml_changing_signature_namespace_prefix_to_default(self):
+    def test_changing_signature_namespace_prefix_to_default(self):
         data = etree.parse(self.example_xml_files[0]).getroot()
         signer = XMLSigner()
         ns = dict()
@@ -389,7 +412,7 @@ class TestSignXML(unittest.TestCase):
         signer.namespaces = ns
         signed = signer.sign(data, key=self.keys["rsa"])
         signed_data = etree.tostring(signed)
-        expected_match = ('<Signature xmlns="%s">') % namespaces["ds"]
+        expected_match = f'<Signature xmlns="{namespaces["ds"]}">'
         self.assertTrue(re.search(expected_match.encode("ascii"), signed_data))
 
     def test_elementtree_compat(self):
@@ -397,38 +420,36 @@ class TestSignXML(unittest.TestCase):
         signer = XMLSigner()
         signer.sign(data, key=self.keys["rsa"])
 
+    saml_test_vectors = [
+        """<samlp:Response xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
+                            xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" ID="responseId">
+            <saml:Assertion xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
+                            xmlns:xs="http://www.w3.org/2001/XMLSchema" ID="assertionId">
+            <ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#" Id="placeholder" />
+            </saml:Assertion>
+            </samlp:Response>""",
+        """<samlp:Response xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
+                            xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" Id="responseId">
+            <saml:Assertion xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
+                            xmlns:xs="http://www.w3.org/2001/XMLSchema" Id="assertionId">
+            <ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#" Id="placeholder" />
+            </saml:Assertion>
+            <saml:Assertion xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
+                            xmlns:xs="http://www.w3.org/2001/XMLSchema" Id="assertion2">
+            </saml:Assertion>
+            </samlp:Response>""",
+    ]
+
     def test_reference_uris_and_custom_key_info(self):
-        with open(os.path.join(os.path.dirname(__file__), "example.pem"), "rb") as fh:
-            crt = fh.read()
-        with open(os.path.join(os.path.dirname(__file__), "example.key"), "rb") as fh:
-            key = fh.read()
+        crt, key = self.load_example_keys()
 
         # Both ID and Id formats. XPath 1 doesn't have case insensitive attribute search
-        for d in [
-            """<samlp:Response xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
-                                     xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" ID="responseId">
-                      <saml:Assertion xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
-                                      xmlns:xs="http://www.w3.org/2001/XMLSchema" ID="assertionId">
-                       <ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#" Id="placeholder" />
-                      </saml:Assertion>
-                     </samlp:Response>""",
-            """<samlp:Response xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"
-                                     xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion" Id="responseId">
-                      <saml:Assertion xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
-                                      xmlns:xs="http://www.w3.org/2001/XMLSchema" Id="assertionId">
-                       <ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#" Id="placeholder" />
-                      </saml:Assertion>
-                      <saml:Assertion xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
-                                      xmlns:xs="http://www.w3.org/2001/XMLSchema" Id="assertion2">
-                      </saml:Assertion>
-                     </samlp:Response>""",
-        ]:
+        for d in self.saml_test_vectors:
             data = etree.fromstring(d)
             reference_uri = ["assertionId", "assertion2"] if "assertion2" in d else "assertionId"
             signed_root = XMLSigner().sign(data, reference_uri=reference_uri, key=key, cert=crt)
-            signed_data_root = XMLVerifier().verify(etree.tostring(signed_root), x509_cert=crt, expect_references=True)[
-                1
-            ]
+            res = XMLVerifier().verify(etree.tostring(signed_root), x509_cert=crt, expect_references=True)
+            signed_data_root = res.signed_xml
             ref = signed_root.xpath(
                 "/samlp:Response/saml:Assertion/ds:Signature/ds:SignedInfo/ds:Reference",
                 namespaces={
@@ -475,6 +496,42 @@ class TestSignXML(unittest.TestCase):
             expect_refs = etree.tostring(s4).decode().count("<ds:Reference")
             XMLVerifier().verify(s4, x509_cert=crt, ignore_ambiguous_key_info=True, expect_references=expect_refs)
 
+    def test_inclusive_namespaces_signing(self):
+        # Test exclusive canonicalization with InclusiveNamespace PrefixList
+        data = etree.fromstring(self.saml_test_vectors[0])
+        reference_uri = "assertionId"
+        signer = XMLSigner(
+            c14n_algorithm=CanonicalizationMethod.EXCLUSIVE_XML_CANONICALIZATION_1_0,
+            signature_algorithm=SignatureMethod.HMAC_SHA256,
+        )
+        sign_args = dict(data=data, reference_uri=reference_uri, key=b"secret")
+        signed_with_insp = signer.sign(inclusive_ns_prefixes=["saml", "samlp", "ds"], **sign_args)
+        self.assertEqual(
+            signed_with_insp.find(".//ds:SignatureValue", namespaces=namespaces).text,
+            "V3HwLySk6TmkPFlqtOU3xCln1Cga7v30rFqy8ZmJITA=",
+        )
+        signed_without_insp = signer.sign(**sign_args)
+        self.assertEqual(
+            signed_without_insp.find(".//ds:SignatureValue", namespaces=namespaces).text,
+            "2Sk3E2SZwQCqX2STDkplw5JJp1ATAB5wsdRdQBAe/7o=",
+        )
+        # Test different c14n methods for signature and payload
+        ref = XMLSignatureReference(
+            URI=reference_uri, c14n_method=CanonicalizationMethod.EXCLUSIVE_XML_CANONICALIZATION_1_0
+        )
+        sign_args = dict(data=data, reference_uri=[ref], key=b"secret")
+        signed = signer.sign(**sign_args)
+        self.assertEqual(
+            signed.find(".//ds:SignatureValue", namespaces=namespaces).text,
+            "2Sk3E2SZwQCqX2STDkplw5JJp1ATAB5wsdRdQBAe/7o=",
+        )
+        sign_args["reference_uri"][0].c14n_method = CanonicalizationMethod.CANONICAL_XML_1_0
+        signed2 = signer.sign(**sign_args)
+        self.assertEqual(
+            signed2.find(".//ds:SignatureValue", namespaces=namespaces).text,
+            "8GPAVJstDxHyuoJqec8C0ssji4zfdXanu1YHGlWbfx0=",
+        )
+
     def test_excision_of_untrusted_comments(self):
         pass  # TODO: test comments excision
 
@@ -513,10 +570,7 @@ class TestSignXML(unittest.TestCase):
         doc = etree.Element("Test", attrib={"Id": "mytest"})
         sigprop = etree.Element("{http://somenamespace}MyCustomProperty")
         sigprop.text = "Some Text"
-        with open(os.path.join(os.path.dirname(__file__), "example.key"), "rb") as file:
-            key = file.read()
-        with open(os.path.join(os.path.dirname(__file__), "example.pem"), "rb") as file:
-            cert = file.read()
+        cert, key = self.load_example_keys()
         signature = XMLSigner(method=methods.detached).sign(
             doc, cert=cert, key=key, reference_uri="#mytest", signature_properties=sigprop
         )
@@ -527,10 +581,7 @@ class TestSignXML(unittest.TestCase):
         doc = etree.Element("{http://somenamespace}Test", attrib={"Id": "mytest"})
         sigprop = etree.Element("{http://somenamespace}MyCustomProperty")
         sigprop.text = "Some Text"
-        with open(os.path.join(os.path.dirname(__file__), "example.key"), "rb") as file:
-            key = file.read()
-        with open(os.path.join(os.path.dirname(__file__), "example.pem"), "rb") as file:
-            cert = file.read()
+        cert, key = self.load_example_keys()
         signer = XMLSigner(method=methods.detached, c14n_algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315")
         signer.namespaces["ns0"] = "http://enveloping.namespace"
         signature = signer.sign(doc, cert=cert, key=key, reference_uri="#mytest", signature_properties=sigprop)
@@ -545,9 +596,84 @@ class TestSignXML(unittest.TestCase):
     def test_payload_c14n(self):
         doc = etree.fromstring('<abc xmlns="http://example.com"><foo xmlns="">bar</foo></abc>')
         self.assertEqual(
-            XMLVerifier()._c14n(doc, algorithm=""), b'<abc xmlns="http://example.com"><foo xmlns="">bar</foo></abc>'
+            XMLSignatureProcessor()._c14n(doc, algorithm=CanonicalizationMethod.CANONICAL_XML_1_1),
+            b'<abc xmlns="http://example.com"><foo xmlns="">bar</foo></abc>',
         )
 
 
+class TestXAdES(unittest.TestCase, LoadExampleKeys):
+    expect_references = {
+        "factura_ejemplo2_32v1.xml": 3,
+        "dss1770.xml": 3,
+        "xades-fake-counter-signature.xml": 3,
+        "Signature-X-SK_DIT-1.xml": 5,
+        "Signature-X-HR_FIN-1.xml": 5,
+        "TEST_S1a_C1a_InTL_VALID.xml": 3,
+        "Signature-X-CZ_SEF-4.xml": 4,
+        "Signature-X-FR_NOT-3.xml": 3,
+        "Signature-X-CZ_SEF-5.xml": 4,
+        "xades-counter-signature-injected.xml": 3,
+        "11068_signed.xml": 3,
+        "signature_property_signed.xml": 3,
+        "nonconformant-Signature-X-ES-100.xml": 3,
+        "nonconformant-Signature-X-ES-103.xml": 3,
+        "nonconformant-dss1770.xml": 3,
+    }
+    signature_policy = XAdESSignaturePolicy(
+        Identifier="urn:sbr:signature-policy:xml:2.0",
+        Description="Test description",
+        DigestMethod=DigestAlgorithm.SHA256,
+        DigestValue="sVHhN1eqNH/PZ1B6h//ehyC1OwRQOrz/tJ3ZYaRrBgA=",
+    )
+    claimed_roles = ["signer"]
+    data_object_format = XAdESDataObjectFormat(Description="Important Document", MimeType="text/xml")
+
+    def test_xades_roundtrip(self):
+        cert, key = self.load_example_keys()
+        with open(os.path.join(os.path.dirname(__file__), "example.xml"), "rb") as fh:
+            doc = etree.parse(fh)
+        signer = XAdESSigner(
+            signature_policy=self.signature_policy,
+            claimed_roles=self.claimed_roles,
+            data_object_format=self.data_object_format,
+        )
+        signed_doc = signer.sign(doc, key=key, cert=cert)
+        verifier = XAdESVerifier()
+        verify_results = verifier.verify(
+            signed_doc, x509_cert=cert, expect_references=3, expect_signature_policy=self.signature_policy
+        )
+        self.assertIsInstance(verify_results[1], XAdESVerifyResult)
+        self.assertTrue(hasattr(verify_results[1], "signed_properties"))
+
+    def test_xades_interop_examples(self):
+        error_conditions = {
+            "altered": InvalidSignature,
+            "newlines": InvalidSignature,
+            "unsupported-signature-algorithm": InvalidInput,
+            "corrupted-cert": etree.DocumentInvalid,  # FIXME - flaky validation
+            "cert-v2-wrong-digest": InvalidDigest,
+            "wrong-sign-cert-digest": InvalidDigest,
+            "nonconformant-X_BE_CONN_10": InvalidDigest,
+            "sigPolStore-noDigest": InvalidInput,
+        }
+        for sig_file in glob(os.path.join(os.path.dirname(__file__), "xades", "*.xml")):
+            print("Verifying", sig_file)
+            with open(sig_file, "rb") as fh:
+                doc = etree.parse(fh)
+            cert = doc.find("//{http://www.w3.org/2000/09/xmldsig#}X509Certificate").text
+            kwargs = dict(x509_cert=cert, expect_references=self.expect_references.get(os.path.basename(sig_file), 2))
+            if "nonconformant" in sig_file:
+                kwargs.update(validate_schema=False)
+            if "sigPolStore" in sig_file:
+                kwargs.update(expect_signature_policy=self.signature_policy)
+            for condition, error in error_conditions.items():
+                if condition in sig_file:
+                    with self.assertRaises(error):
+                        XAdESVerifier().verify(doc, **kwargs)
+                    break
+            else:
+                XAdESVerifier().verify(doc, **kwargs)
+
+
 if __name__ == "__main__":
     unittest.main()
